from aegon.libutils import sort_by_energy
from aegon.libroulette import get_fitness
#-------------------------------------------------------------------------------
def display_mol_info(moleculein, flagsum=1):
    if len(moleculein)==0:
        print("\n------------ALL MOLECULES DISCRIMINATED. GLOMOS FINISH.------------")
    molzz=sort_by_energy(moleculein, 1)
    for ii, imol in enumerate(molzz):
        deltae=imol.info['e'] - molzz[0].info['e']
        jj=str(ii+1).zfill(5)
        if flagsum == 1:
            fitness=get_fitness(moleculein)
            print("#%s %-12s %.6f kcal/mol (%.6f) (f=%.2f)" %(jj, imol.info['i'], imol.info['e'], deltae, fitness[ii]))
        else:
            print("#%s %-12s %.6f kcal/mol (%.6f)" %(jj, imol.info['i'], imol.info['e'], deltae))
#-------------------------------------------------------------------------------



import time
import warnings
warnings.filterwarnings("ignore", message="cuaev not installed")
import os
import sys
import torchani
import queue
import contextlib
from multiprocessing import Process, Queue
from ase.optimize import BFGS
from aegon.libutils import readxyzs, writexyzs
#-------------------------------------------------------------------------------
eVtokcalpermol = 23.060548012069496
#-------------------------------------------------------------------------------
@contextlib.contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
#-------------------------------------------------------------------------------
def ANI_single(atoms, opt='ANI1ccx', preclist=[1E-03, 1E-04, 1E-05]):
    timein=time.strftime("%c")
    print('%s at %s' %(atoms.info['i'], timein))
    moleculeout=atoms.copy()
    for prec in preclist:
        with suppress_stdout():
            calculator = {
                'ANI1x':   torchani.models.ANI1x().ase(),
                'ANI1ccx': torchani.models.ANI1ccx().ase(),
                'ANI2x':   torchani.models.ANI2x().ase()
            }[opt]
        moleculeout.calc = calculator
        dyn = BFGS(moleculeout, logfile=None)
        dyn.run(fmax=prec, steps=200)
    energy = moleculeout.get_potential_energy()
    moleculeout.info['e'] = energy * eVtokcalpermol
    return moleculeout
#-------------------------------------------------------------------------------
def ANI_single_to_file(atoms, outname, opt='ANI1ccx', preclist=[1E-03, 1E-04, 1E-05]):
    moleculeout= ANI_single(atoms, opt, preclist)
    writexyzs(moleculeout, outname)
#-------------------------------------------------------------------------------
def do_job(tasks_to_accomplish, tasks_that_are_done, opt='ANI1ccx', preclist=[1E-03, 1E-04, 1E-05]):
    while True:
        try:
            task = tasks_to_accomplish.get_nowait()
        except queue.Empty:
            break
        else:
            mol = task
            outname = f"{mol.info['i']}_opt.xyz"
            ANI_single_to_file(mol, outname, opt, preclist)
            tasks_that_are_done.put(outname)
#-------------------------------------------------------------------------------
def ANI(moleculelist, n_jobs=1, opt='ANI1ccx', preclist=[1E-03, 1E-04, 1E-05]):
    tasks_to_accomplish = Queue()
    tasks_that_are_done = Queue()
    processes = []
    for mol in moleculelist:
        tasks_to_accomplish.put(mol)
    for _ in range(n_jobs):
        p = Process(target=do_job, args=(tasks_to_accomplish, tasks_that_are_done, opt, preclist))
        processes.append(p)
        p.start()
    for p in processes:
        p.join()
    optimized_molecules = []
    while not tasks_that_are_done.empty():
        xyzfile = tasks_that_are_done.get()
        mol = readxyzs(xyzfile)[0]
        mol.info['c']=1
        optimized_molecules.append(mol)
        os.remove(xyzfile)
    return optimized_molecules
#-------------------------------------------------------------------------------
#from joblib import Parallel, delayed
#def ANI_parallel_without_queuing(mol_list, n_jobs = 1, opt='ANI1ccx', preclist=[1E-03, 1E-04, 1E-05]):
#    results = Parallel(n_jobs = n_jobs)(delayed(ANI)(mol, opt, preclist) for mol in mol_list)
#    return results
#-------------------------------------------------------------------------------


from ase.io                import read
from aegon.libutils        import writexyzs, sort_by_energy, cutter_energy, rename, adjacency_matrix
from aegon.libstdio        import read_main_input
from aegon.librotamers     import get_bridge_left_right, make_random_rotamers, make_crossover_rotamers, make_mutant_rotamers
from aegon.libsel_roulette import get_fitness, get_roulette_wheel_selection
from aegon.libdiscusr      import comparator_usr_serial, molin_sim_molref
from glomos.libcalc_ani    import ANI
from glomos.libtools       import display_mol_info
ndigit1=3
ndigit2=4
preclist=[1E-03, 1E-05]
#-------------------------------------------------------------------------------
def conformational(inputfile='INPUT.txt'):
    #Reading variables
    df = read_main_input(inputfile)
    rotamer_seed=df.get_str(key='rotamer_seed', default='rotamer.xyz')
    nof_initpop=df.get_int(key='nof_initpop', default=10)
    nof_matings=df.get_int(key='nof_matings', default=5)
    nof_mutants=df.get_int(key='nof_mutants', default=5)
    tol_similarity=df.get_float(key='tol_similarity', default=0.95)
    cutoff_energy=df.get_float(key='cutoff_energy', default=5.0)
    cutoff_population=df.get_int(key='cutoff_population', default=8)
    nof_generations=df.get_int(key='nof_generations', default=3)
    nof_repeats=df.get_int(key='nof_repeats', default=2)
    nof_stagnant=df.get_int(key='nof_stagnant', default=3)
    calculator=df.get_str(key='calculator', default='ANI1ccx')
    nof_processes=df.get_int(key='nof_processes', default=1)
    atoms = read(rotamer_seed, format='xyz') 
    atoms.info['e']=0.0
    atoms.info['i']='Rotamer'
    #Welcome
    adjmatrix = adjacency_matrix(atoms)
    bridgelist = get_bridge_left_right(adjmatrix)
    print('----------------- Genetic Algorithm for Rotamers -----------------')
    print('Chemical Formula      = %s'    %(atoms.get_chemical_formula()))
    if ( len(bridgelist) == 0 ):
        print('Selected XYZ Mol File has not bridges.')
        print('Choose (as seed) a molecule with at least a rotable bond.')
        exit()
    print("Number of bridges     = %d" %(len(bridgelist)))
    for ibridge in range(len(bridgelist)):
        print(' -Bridge[%d]           = (%d, %d)' %(ibridge+1, bridgelist[ibridge][0]+1, bridgelist[ibridge][1]+1))
    print('\nEVOLUTIVE PARAMETERS:')
    print('Initial Population    = %d'    %(nof_initpop))
    print('Number of matings     = %d'    %(nof_matings))
    print('Number of mutants     = %d'    %(nof_mutants))
    print('\nDISCRIMINATION PARAMETERS:')
    print('Tol for similarity    = %4.2f' %(tol_similarity))
    print('Energy Cut-off        = %.2f'  %(cutoff_energy))
    print('Max population size   = %d'    %(cutoff_population))
    print('\nSTOP CRITERION:')
    print('Max generations       = %d'    %(nof_generations))
    print('Max repeated rotamers = %d'    %(nof_repeats))
    print('Max stagnant cycles   = %d'    %(nof_stagnant))
    print()
    print('Theory Level          = %s'    %(calculator))
    #Main Algorithm
    print('---------------------------GENERATION 0---------------------------')
    print('Construction of the guest rotamers. Initial seed (random_000_001) as reference\n')
    xrand=make_random_rotamers(atoms, nof_initpop, bridgelist, adjmatrix)
    rename(xrand, 'random_'+str(0).zfill(ndigit1), ndigit2)
    print('Optimization at %s:' %(calculator))
    xopt=ANI(xrand, n_jobs=nof_processes, opt=calculator, preclist=preclist)
    print('')
    print('Discrimination. Max population size=%d; Energy Cut-off=%.2f; Tol for similarity=%4.2f' %(cutoff_population,cutoff_energy,tol_similarity))
    xopt=cutter_energy(xopt, cutoff_energy)
    xopt_sort=sort_by_energy(xopt, 1)
    xopt_sort=comparator_usr_serial(xopt_sort, tol_similarity)
    xopt_sort=xopt_sort[:cutoff_population]
    print('\nGLOBAL SUMMARY')
    display_mol_info(xopt_sort)
    writexyzs(xopt_sort, 'summary.xyz')
    namesi=[imol.info['i'] for imol in xopt_sort][:nof_repeats]
    count=0
    for igen in range(nof_generations):
        print("\n---------------------------GENERATION %d---------------------------" %(igen+1))
        print('Construction of crossovers ...')
        list_p=get_roulette_wheel_selection(xopt_sort, nof_matings)
        list_m=get_roulette_wheel_selection(xopt_sort, nof_matings)
        atoms_list_out=make_crossover_rotamers(list_p, list_m, bridgelist, adjmatrix)
        rename(atoms_list_out, 'mating_'+str(igen+1).zfill(ndigit1), ndigit2)
        print('Construction of mutants    ...\n')
        atoms_list_mut=make_mutant_rotamers(xopt_sort[:nof_mutants], bridgelist, adjmatrix)
        rename(atoms_list_mut, 'mutant_'+str(igen+1).zfill(ndigit1), ndigit2)
        print('Optimization at %s:' %(calculator))
        generation_opt=ANI(atoms_list_out+atoms_list_mut, n_jobs=nof_processes, opt=calculator, preclist=preclist)
        print('\nDiscrimination. Max population size=%d; Energy Cut-off=%.2f; Tol for similarity=%4.2f' %(cutoff_population,cutoff_energy,tol_similarity))
        generation_opt=cutter_energy(generation_opt, cutoff_energy)
        generation_opt=sort_by_energy(generation_opt,1)
        generation_opt=comparator_usr_serial(generation_opt, tol_similarity)
        generation_opt=molin_sim_molref(generation_opt, xopt_sort, tol_similarity)
        xopt_sort=sort_by_energy(xopt_sort+generation_opt, 1)
        xopt_sort=xopt_sort[:cutoff_population]
        print('\nGLOBAL SUMMARY')
        display_mol_info(xopt_sort)
        writexyzs(xopt_sort, 'summary.xyz')
        namesj=[imol.info['i'] for imol in xopt_sort][:nof_repeats]
        numij=[1 for i, j in zip(namesi,namesj) if i == j]
        count=count+1 if sum(numij) == nof_repeats else 1
        if count == nof_stagnant:
            print("\nEarly termination. Max repeated isomers (%d) reached at the Max stagnant cycles (%d)." %(nof_repeats, nof_stagnant))
            break
        namesi=namesj
    print("\nGlobal optimization complete.")
#-------------------------------------------------------------------------------



from aegon.libutils import sort_by_energy, cutter_energy, rename
from aegon.libmolgen  import make_molecules_random
from aegon.libcrossover import crossover_deavenho
from aegon.libstdio import read_main_input
#from aegon.libcalc_emt import ene_EMT, opt_EMT
from aegon.libcalc_lj import opt_LJ_parallel
from aegon.libsel_roulette import get_fitness, get_roulette_wheel_selection
from aegon.libdiscusr import comparator_usr_serial, molin_sim_molref
#------------------------------------------------------------------------------------------
def display_mol_info(moleculein, flagsum=0, ngen=0):
    if len(moleculein)==0:
        print("\n------------ALL MOLECULES DISCRIMINATED. GLOMOS FINISH.------------")
    if flagsum == 0:
        print("---------------------------GENERATION %d---------------------------" %(ngen))
    else:
        print("--------------------------GLOBAL SUMMARY---------------------------")
    molzz=sort_by_energy(moleculein, 1)
    fitness=get_fitness(moleculein)
    for ii, imol in enumerate(molzz):
        deltae=imol.info['e'] - molzz[0].info['e']
        jj=str(ii+1).zfill(5)
        if flagsum == 0:
            print("#%s %-14s with %13.8f eV (%10.8f)" %(jj, imol.info['i'], imol.info['e'], deltae))
        else:
            print("#%s %-14s with %13.8f eV (%10.8f) (f=%3.2f)" %(jj, imol.info['i'], imol.info['e'], deltae, fitness[ii]))
    if flagsum != 0: print('')
#------------------------------------------------------------------------------------------
def genetic_algorithm(inputfile):
    df = read_main_input(inputfile)
    composition = df.get_comp(key='COMPOSITION')
    atomlist=composition.atoms
    nameid=composition.name
    mono=True if len(composition.elements)==1 else False
    ninitpop=df.get_int(key='nof_initpop', default=10)
    nmatings=df.get_int(key='nof_matings', default=5)
    ngenerations=df.get_int(key='nof_generations', default=5)
    tolsim=df.get_float(key='tol_similarity', default=0.95)
    ecut=df.get_float(key='cutoff_energy', default=10.0)
    npop=df.get_int(key='cutoff_population', default=10)
    nproc=df.get_int(key='nof_processes', default=2)
    print('-------- Global Optimization of Molecular Systems (GLOMOS) --------')
    print('Input file           = %s' %(inputfile))
    print('Option               = Modified Genetic Algorithm')
    print('Composition          = %s' %(nameid))
    print('Initial Population   = %d' %(ninitpop))
    print('Tol for similarity   = %4.2f' %(tolsim))
    print('Energy Cut-off       = %f' %(ecut))
    print('Max population size  = %d\n' %(npop))
    xrand = make_molecules_random(atomlist, ninitpop)
    xrand=rename(xrand,'random_00',4)
    #xopt=[opt_EMT(ix) for ix in xrand]
    xopt=opt_LJ_parallel(xrand, nproc)
    xopt=cutter_energy(xopt, ecut)
    xopt_sort=sort_by_energy(xopt, 1)
    xopt_sort=comparator_usr_serial(xopt_sort, tolsim, mono)
    xopt_sort=xopt_sort[:npop]
    display_mol_info(xopt_sort, flagsum=0, ngen=0)
    print('')
    for igen in range(ngenerations):
        list_p=get_roulette_wheel_selection(xopt_sort, nmatings)
        list_m=get_roulette_wheel_selection(xopt_sort, nmatings)
        atoms_list_out=[]
        for i in range(nmatings):
            cross = crossover_deavenho(list_p[i], list_m[i], atomlist)
            if cross: atoms_list_out.extend([cross])
        atoms_list_out=rename(atoms_list_out,'mating_'+str(igen+1).zfill(2),4)
        #cross_opt=[opt_EMT(ix) for ix in atoms_list_out]
        cross_opt=opt_LJ_parallel(atoms_list_out, nproc)
        cross_opt=cutter_energy(cross_opt, ecut)
        cross_opt=sort_by_energy(cross_opt,1)
        cross_opt=comparator_usr_serial(cross_opt, tolsim, mono)
        display_mol_info(cross_opt, flagsum=0, ngen=igen+1)
        cross_opt=molin_sim_molref(cross_opt, xopt_sort, tolsim, mono)
        xopt_sort=sort_by_energy(xopt_sort+cross_opt, 1)
        xopt_sort=xopt_sort[:npop]
        display_mol_info(xopt_sort, flagsum=1)
    return xopt_sort
#------------------------------------------------------------------------------------------
